{
  Multi-Check

  To use different 'check.exe' programs depending on actual type of
  game with braindead autohosts which don't have this configurable.
  Can also be used to multiplex other host programs.
}
{$A+,B-,D+,E+,F-,G+,I+,L+,N-,O-,P-,Q-,R-,S+,T+,V-,X+,Y+}
{$M 8192,0,0}
PROGRAM MCheck;

USES DOS;

CONST DIRSEP = [':','\','/'];
      ALL_PLAYERS = $FFE;   { 1 shl 1 .. 1 shl 11 }
      VERSION = '0.2';

VAR players     : WORD;                            { bitfield of players, 1 shl number }
    vars        : ARRAY['A'..'Z'] OF STRING;       { variables }
    config_file : STRING;                          { config file name }
    mode        : (NoMode, AllMode, SingleMode);   { how to run the command }
    command     : STRING;                          { command }
    norun_mode  : BOOLEAN;                         { /n }
    test_mode   : BOOLEAN;                         { /t }
    line        : INTEGER;                         { line number; zero if not reading config file }
    exitcode    : INTEGER;                         { exit code of programs }

    output_redir : BOOLEAN;                        { true if output redirection is enabled }
    output_fd    : INTEGER;                        { fd of the output file }
    stdout_fd    : INTEGER;                        { copy of original stdout }
    stderr_fd    : INTEGER;                        { copy of original stderr }

{ Exit with error message }
PROCEDURE Die(CONST s:STRING);
BEGIN
  IF line<>0 THEN Writeln('mcheck: config file line ', line, ': ', s)
             ELSE Writeln('mcheck: ', s);
  Halt(255);
END;

{ Initialize: parse parameters and set up variables }
PROCEDURE Init;
VAR c       : CHAR;
    i, j, k : INTEGER;
    s       : STRING;
BEGIN
  FOR c:='A' TO 'Z' DO vars[c] := '';
  players := 0;
  norun_mode := FALSE;
  test_mode := FALSE;
  output_redir := FALSE;
  line := 0;

  IF ParamCount=0 THEN BEGIN
    Writeln('MCheck v' + VERSION + ' - CHECK.EXE Multiplexer - (c) 2003 Stefan Reuther');
    Writeln;
    Writeln('Usage: MCHECK [<gamedir> [<rootdir>]] [/<player>] [/t] [/n]');
    Writeln;
    Writeln('  /t    only test configuration');
    Writeln('  /n    do not run commands, just print them');
    Halt;
  END;

  FOR i:=1 TO ParamCount DO BEGIN
    s := ParamStr(i);
    Val(s, j, k);
    IF (k=0) AND (j>0) AND (j<=11) THEN BEGIN
      { player number }
      players := players OR (1 SHL j);
    END ELSE IF (S[1]='/') THEN BEGIN
      { set of players or options }
      FOR j:=2 TO Length(s) DO BEGIN
        CASE s[j] OF
          '1'..'9': players := players OR (1 SHL (Ord(s[j])-Ord('0')));
          'A','a':  players := players OR (1 SHL 10);
          'B','b':  players := players OR (1 SHL 11);
          '*':      players := ALL_PLAYERS;
          'N','n':  norun_mode := TRUE;
          'T','t':  test_mode := TRUE;
        ELSE Die('Invalid option "' + s[j] + '"');
        END;
      END;
    END ELSE IF (vars['G']='') THEN BEGIN
      { game directory }
      vars['G'] := s;
    END ELSE IF (vars['R']='') THEN BEGIN
      { root directory }
      vars['R'] := s;
    END ELSE BEGIN
      Die('Too many parameters');
    END;
  END;

  { get default root directory }
  IF vars['R']='' THEN BEGIN
    s := ParamStr(0);
    vars['R'] := '.';
    FOR i:=Length(s) DOWNTO 1 DO
     IF s[i] IN DIRSEP THEN BEGIN
       s[0] := Chr(i-1);
       vars['R']:=s;
       Break;
     END;
  END;

  { get default game directory }
  IF vars['G']='' THEN vars['G']:='.';

  { shell name }
  vars['S'] := GetEnv('COMSPEC') + ' /c';

  { get config file name }
  s := ParamStr(0);
  config_file := s + '.ini';
  FOR i:=Length(s) DOWNTO 1 DO BEGIN
    IF s[i]='.' THEN BEGIN
      config_file := Copy(s, 1, i) + 'ini';
    END ELSE IF s[i] IN DIRSEP THEN Break;
  END;
END;

{ Trim leading/trailing whitespace }
PROCEDURE Trim(VAR s:STRING);
BEGIN
  WHILE (s<>'') AND (s[1]=' ') DO Delete(s, 1, 1);
  WHILE (s<>'') AND (s[Length(s)]=' ') DO Dec(s[0]);
END;

{ convert to uppercase }
PROCEDURE Up(VAR s:STRING);
VAR i:INTEGER;
BEGIN
  FOR i:=1 TO Length(s) DO s[i] := Upcase(s[i]);
END;

{ convert tabs to spaces }
PROCEDURE Tabs(VAR s:STRING);
VAR i:INTEGER;
BEGIN
  FOR i:=1 TO Length(s) DO IF s[i]=#9 THEN s[i]:=' ';
END;

{ Split string s at delimiter delim. Somehow like C's strtok. Also eats
  whitespace }
PROCEDURE Eat(VAR s, out:STRING; CONST delim:STRING);
VAR i:INTEGER;
BEGIN
  Trim(s);
  i := Pos(delim, s);
  IF i>0 THEN BEGIN
    out := Copy(s, 1, i-1);
    Delete(s, 1, i+Length(delim)-1);
  END ELSE BEGIN
    out := s;
    s := '';
  END;
  Trim(out);
  Trim(s);
END;

{ Expand variables }
FUNCTION Expand(CONST s:STRING):STRING;
VAR out : STRING;
    i   : INTEGER;
    c   : CHAR;
BEGIN
  i := 1;
  out := '';
  WHILE i <= Length(s) DO BEGIN
    IF (s[i]='%') THEN BEGIN
      Inc(i);
      IF i > Length(s) THEN Die('Lone % in string');
      c := Upcase(s[i]);
      IF c='%' THEN out := out + c ELSE
      IF (c>='A') AND (c<='Z') THEN BEGIN
        out := out + vars[c];
        IF (vars[c]<>'') AND (vars[c][Length(vars[c])] IN ['/','\'])
         AND (i < Length(s)) AND (s[i+1] IN ['/','\']) THEN BEGIN
          Inc(i);
        END;
      END ELSE Die('Invalid % escape');
    END ELSE BEGIN
      out := out + s[i];
    END;
    Inc(i);
  END;
  Expand := out;
END;

{ Do one test. Verb is the test command, test is its argument. }
FUNCTION DoTest(VAR verb, test:STRING):BOOLEAN;
VAR F:FILE;
    i, j:INTEGER;
BEGIN
  IF verb='IF_EXISTS' THEN BEGIN
    Assign(F, Expand(test));
    {$I-}Reset(F, 1);{$I+}
    IF IOResult=0 THEN BEGIN
      Close(F);
      DoTest:=TRUE;
    END ELSE DoTest:=FALSE;
  END ELSE IF verb='IF_DIR' THEN BEGIN
    Up(test);

    verb := vars['G'];  { hack hack hack }
    Up(verb);
    j:=0;
    FOR i:=1 TO Length(verb)-1 DO IF verb[i] IN DIRSEP THEN j:=i;
    Delete(verb, 1, j);
    DoTest := (verb = test);
  END ELSE IF verb='IF_PLAYERS' THEN BEGIN
    Expand(test);
    j := 0;
    FOR i:=1 TO Length(test) DO CASE test[i] OF
      ' ':;
      '1'..'9': j := j OR (1 SHL (Ord(test[i])-Ord('0')));
      'A','a':  j := j OR (1 SHL 10);
      'B','b':  j := j OR (1 SHL 11);
      '*':      j := ALL_PLAYERS;
    END;
    DoTest := (players AND j) = j;
  END ELSE IF verb='ALWAYS' THEN BEGIN
    IF test<>'' THEN Die('"Always" does not accept parameters');
    DoTest := TRUE;
  END ELSE IF verb='NEVER' THEN BEGIN
    IF test<>'' THEN Die('"Never" does not accept parameters');
    DoTest := FALSE;
  END ELSE BEGIN
    Die('Invalid test keyword "' + verb + '"');
  END;
END;

{ Read one line, and postprocess it }
PROCEDURE ReadLine(VAR f:TEXT; VAR s:STRING);
BEGIN
  Readln(F, S); Tabs(S); Trim(S);
  Inc(line);
END;

{ Load configuration file }
PROCEDURE LoadConfig;
VAR F, F2 : TEXT;
    S     : STRING;
    verb  : STRING;
    test  : STRING;
    test_ok : BOOLEAN;
    c       : CHAR;
BEGIN
  Assign(F, config_file);
  {$I-}Reset(F);{$I+}
  IF IOResult<>0 THEN Die('Unable to open config file "' + config_file + '"');
  IF test_mode OR norun_mode THEN Writeln('mcheck: using config file "', config_file, '"');
  Mode := NoMode;
  WHILE NOT EOF(F) DO BEGIN
    ReadLine(F, S);
    IF (S='') OR (S[1]=';') THEN Continue;

    Eat(S, verb, ' ');
    Up(verb);
    IF verb='RUN_EACH' THEN BEGIN
      Mode := SingleMode;
      Expand(S);         { to test whether it's valid }
      command := S;
    END ELSE IF verb='RUN_ONCE' THEN BEGIN
      Mode := AllMode;
      Expand(S);         { to test whether it's valid }
      command := S;
    END ELSE IF verb='SET' THEN BEGIN
      Eat(S, verb, '=');
      IF (verb<>'') AND (verb[1]='%') THEN Delete(verb, 1, 1);
      IF (Length(verb)<>1) OR NOT (Upcase(verb[1]) IN ['A'..'Z']) THEN Die('Invalid variable name');
      vars[Upcase(verb[1])] := Expand(S);
    END ELSE IF verb='SET_FILE' THEN BEGIN
      Eat(S, verb, '=');
      IF (verb<>'') AND (verb[1]='%') THEN Delete(verb, 1, 1);
      IF (Length(verb)<>1) OR NOT (Upcase(verb[1]) IN ['A'..'Z']) THEN Die('Invalid variable name');
      Assign(F2, Expand(S));
      {$I-}Reset(F2);{$I+}
      IF IOResult=0 THEN BEGIN
        Readln(F2, vars[Upcase(verb[1])]);
        Close(F2);
      END ELSE vars[Upcase(verb[1])] := '';
    END ELSE BEGIN
      IF Mode=NoMode THEN Die('Expected a "run_XXX" command first');
      test_ok := TRUE;
      WHILE Pos('&', S)<>0 DO BEGIN
        Eat(S, test, '&');
        IF NOT DoTest(verb, test) THEN test_ok := FALSE;
        WHILE S='' DO BEGIN
          { if S is empty, we're still expecting more tests on the
            following lines. }
          IF EOF(F) THEN Die('Unexpected end of file');
          ReadLine(F, S);
          IF (S='') OR (S[1]=';') THEN S:='';
        END;
        Eat(S, verb, ' ');
        Up(verb);
      END;
      IF NOT DoTest(verb, S) THEN test_ok := FALSE;
      IF test_ok AND NOT test_mode THEN BEGIN
        { exit here only if we're not in test mode. In test mode,
          we parse the whole file. }
        line := 0;
        Close(F);
        Exit;
      END;
    END;
  END;
  Close(F);
  line := 0;
  IF test_mode THEN BEGIN
    Writeln('Config file looks ok. Here are the final variable values:');
    FOR c:='A' TO 'Z' DO
     IF vars[c]<>'' THEN Writeln('  ', c, ' = ', vars[c]);
    Halt;
  END ELSE BEGIN
    Die('Unable to find matching rule for these parameters.');
  END;
END;

{ Binding for 'dup' syscall (copy a fd to a new fd) }
FUNCTION Dup(fd:INTEGER):INTEGER; ASSEMBLER;
ASM
        mov     bx, fd
        mov     ah, 45h
        int     21h
        jnc     @ok
        mov     ax, -1
  @ok:
END;

{ Binding for 'dup2' syscall (make fd /dest/ refer to file /src/) }
FUNCTION DupTo(dest, src:INTEGER):INTEGER; ASSEMBLER;
ASM
        mov     bx, src
        mov     cx, dest
        mov     ah, 46h
        int     21h
END;

{ Run one command }
PROCEDURE Run(s:STRING);
VAR x:STRING;
    i:INTEGER;
BEGIN
  Writeln('$ ', s);
  Flush(Output);
  IF NOT norun_mode THEN BEGIN
    Eat(s, x, ' ');
    SwapVectors;
    IF output_redir THEN BEGIN
      DupTo(1, output_fd);
      DupTo(2, output_fd);
    END;
    Exec(x, s);
    SwapVectors;
    IF DOSError<>0 THEN BEGIN
      Writeln('DOS Error ', DOSError, ' while running "', x, '".');
      Flush(Output);
    END ELSE BEGIN
      i := DOSExitCode;
      IF i<>0 THEN Writeln('*** Command returned ', i, ' exit status ***');
      Flush(Output);
      exitcode := exitcode OR i;
    END;
    IF output_redir THEN BEGIN
      DupTo(1, stdout_fd);
      DupTo(2, stderr_fd);
    END;
  END;
END;

{ Prepare output redirection. This prepares the xxx_fd variables and sets
  output_redir if we want redirection. This also removes the actual
  redirection from the command line. }
PROCEDURE PrepareOutputRedirection;
VAR i, j   : INTEGER;
    append : BOOLEAN;
    name   : STRING;
    F      : FILE;
    F_fd   : INTEGER ABSOLUTE F;
LABEL File_is_open;
BEGIN
  i := Pos('>', command);
  IF i=0 THEN Exit;
  j := i+1;
  IF (j <= Length(command)) AND (command[j] = '>') THEN Inc(j);
  WHILE (j <= Length(command)) AND (command[j] = ' ') DO Inc(j);
  WHILE (j <= Length(command)) AND (command[j] <> ' ') DO Inc(j);

  { j now points after the last character of the redirection file name }
  name := Copy(command, i+1, j-i-1);
  Delete(command, i, j-i);
  Trim(command);
  IF (name<>'') AND (name[1]='>') THEN BEGIN
    append := TRUE;
    Delete(name, 1, 1);
  END ELSE BEGIN
    append := FALSE;
  END;
  name := Expand(name);
  Trim(name);
  IF name='' THEN Die('Invalid output redirection');

  IF norun_mode THEN BEGIN
    IF append THEN Writeln('Output will be appended to "', name, '"')
              ELSE Writeln('Output will be written to "', name, '"');
    Exit;
  END;

  Assign(F, name);
  IF append THEN BEGIN
    FileMode := 2;
    {$I-}Reset(F, 1);{$I+}
    IF IOResult=0 THEN BEGIN
      Seek(F, FileSize(F));
      GOTO File_is_open;
    END;
  END;
  {$I-}Rewrite(F, 1);{$I+}
  IF IOResult<>0 THEN Die('Unable to open file "' + name + '" for output redirection.');
 File_is_open:
  output_redir := TRUE;
  output_fd    := F_fd;
  stdout_fd    := Dup(1);
  stderr_fd    := Dup(2);
  IF (stdout_fd < 0) OR (stderr_fd < 0) THEN Die('Unable to duplicate standard handles - too many open files?');
END;

CONST elf : ARRAY[1..11] OF CHAR='123456789AB';

VAR i : INTEGER;
BEGIN
  Init;
  IF players=0 THEN players:=ALL_PLAYERS;
  LoadConfig;
  PrepareOutputRedirection;
  exitcode := 0;
  IF Mode=AllMode THEN BEGIN
    vars['P'] := '';
    FOR i:=1 TO 11 DO IF players AND (1 SHL i)<>0 THEN vars['P'] := vars['P'] + elf[i];
    Run(Expand(command));
  END ELSE BEGIN
    FOR i:=1 TO 11 DO IF players AND (1 SHL i)<>0 THEN BEGIN
      vars['P'] := elf[i];
      Str(i, vars['D']);
      Run(Expand(command));
    END;
  END;
  Halt(exitcode);
END.
